/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is gContactSync.
 *
 * The Initial Developer of the Original Code is
 * Josh Geenen <gcontactsync@pirules.net>.
 * Portions created by the Initial Developer are Copyright (C) 2008
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */
/**
 * oAddressBook
 * A class for a Thunderbird Address Book with methods to add, modify, obtain, 
 * and delete cards.
 * @param aDirectory The actual directory.
 * @constructor
 * @class
 */
function oAddressBook(aDirectory) {
  this.mDirectory = aDirectory;
  // make sure the directory is valid
  if (!this.isDirectoryValid(this.mDirectory))
    throw "Invalid directory supplied to the oAddressBook constructor" +
          "\nCalled by: " + this.caller +
          oStringBundle.getStr("pleaseReport");
  // get the directory's URI
  if (this.mDirectory.URI)
    this.mURI = this.mDirectory.URI;
  else {
    this.mDirectory.QueryInterface(Ci.nsIAbMDBDirectory);
    this.mURI = this.mDirectory.getDirUri();
  }
}

oAddressBook.prototype = {
  mURI: {}, // the uniform resource identifier of the directory
  mCards: [], // the cards within this address book
  mCardsUpdate: false, // set to true when mCards should be updated
  /**
   * oAddressBook.addCard
   * Adds the card to this address book and returns the added card.
   * @param aCard The card to add.
   * @return An MDB
   */
  addCard: function oAddressBook_addCard(aCard) {
    oAbManager.checkCard(aCard); // check the card's validity first
    try {
      this.mCards.push(aCard);
      return this.mDirectory.addCard(aCard); // then add it and return the MDBCard
    }
    catch(e) {
      oLOGGER.LOG_ERROR("Unable to add card to the directory with URI: " +
                       this.URI, e);
    }
  },
  /**
   * oAddressBook.getAllCards
   * Returns an array of all of the cards in this Address Book.
   * @return  An array of the nsIAbCards in this Address Book.
   */
  getAllCards: function oAddressBook_getAllCards() {
    this.mCards = [];
    var iter = this.mDirectory.childCards;
    if (oAbManager.mVersion == 3) { // TB 3
      while (iter.hasMoreElements()) {
        data = iter.getNext();
        if (data instanceof nsIAbCard && !data.isMailList)
          this.mCards.push(data);
      }
    }
    else { // TB 2
      // use nsIEnumerator...
      try {
        iter.first();
        do {
          var data = iter.currentItem();
          if(data instanceof nsIAbCard && !data.isMailList)
            this.mCards.push(data);
          iter.next();
        } while (Components.lastResult == 0);
      } catch(e) {} // error is expected when finished   
    }
    return this.mCards;
  },
  /**
   * oAddressBook.getAllLists
   * Returns an an object containing oMailList objects whose attribute name is
   * the name of the mail list.
   * @param skipGetCards True to skip getting the cards of each list.
   * @return An object containing oMailList objects.
   */
  getAllLists: function oAddressBook_getAllLists(skipGetCards) {
    // same in Thunderbird 2 and 3
    var iter = this.mDirectory.childNodes;
    var obj = {};
    while (iter.hasMoreElements()) {
      data = iter.getNext();
      if (data instanceof Ci.nsIAbDirectory && data.isMailList) {
        var list = new oMailList(data, this, skipGetCards);
        var id = list.getGroupID();
        obj[id] = list;
        oLOGGER.VERBOSE_LOG("Found a list named - " + list.getName() + " with" +
                           " an ID - " + id);
      }
    }
    return obj;
  },
  /**
   * oAddressBook.getListByDesc
   * Finds and returns the first Mail List that matches the given nickname in
   * this address book.
   * @param aNickName The nickname to search for.  If null then this
   *                  function returns nothing.
   * @return A new oMailList object containing a list that matches the
   *         nickname or nothing if the list wasn't found.
   */
  getListByNickName: function oAddressBook_getListByNickName(aNickName) {
    if (!aNickName)
      return;
    // same in Thunderbird 2 and 3
    var iter = this.mDirectory.childNodes;
    while (iter.hasMoreElements()) {
      data = iter.getNext();
      if (data instanceof Ci.nsIAbDirectory && data.isMailList &&
          data.listNickName == aNickName) {
        var list = new oMailList(data, this, true);
        return list;
      }
    }
  },
  /**
   * oAddressBook.addList
   * Creates a new mail list, adds it to the address book, and returns a
   * oMailList object containing the list.
   * @param aName     The new name for the mail list.
   * @param aNickName The nickname for the mail list.
   * @return A new oMailList object containing the newly-made Mail List with the
   *         given name and nickname.
   */
  addList: function oAddressBook_addList(aName, aNickName) {
    if (!aName)
      throw "Error - aName sent to addList is invalid";
    if (!aNickName)
      throw "Error - aNickName sent to addList is invalid";
    var list = Cc["@mozilla.org/addressbook/directoryproperty;1"]
                .createInstance(Ci.nsIAbDirectory);
    list.dirName = aName;
    list.listNickName = aNickName;
    list.isMailList = true;
    this.mDirectory.addMailList(list);
    // list can't be QI'd to an MDBDirectory, so the new list has to be found...
    var realList = this.getListByNickName(aNickName);
    return realList;
  },
  /**
   * oAddressBook.deleteCards
   * Deletes the nsIAbCard from the nsIAbDirectory Address Book.  If the card
   * isn't in the book nothing will happen
   * @param aCard   The card to delete from the directory
   */
  deleteCards: function oAddressBook_deleteCards(aCards) {
    if (!(aCards && aCards.length && aCards.length > 0))
      return;
    var arr;
    if (oAbManager.mVersion == 3) { // TB 3
      arr = Cc["@mozilla.org/array;1"].createInstance(Ci.nsIMutableArray);
      for (var i = 0; i < aCards.length; i++) {
        oAbManager.checkCard(aCards[i]);
        arr.appendElement(aCards[i], false);
      }
    }
    else { // TB 2
      arr =  Cc["@mozilla.org/supports-array;1"]
              .createInstance(Ci.nsISupportsArray);
      for (var i = 0; i < aCards.length; i++) {
        oAbManager.checkCard(aCards[i]);
        arr.AppendElement(aCards[i], false);
      }
    }
    if (arr) { // make sure arr isn't null (mailnews bug 448165)
      this.mCardsUpdate = true;
      this.mDirectory.deleteCards(arr);
    }
  },
  /**
   * oAddressBook.updateCard
   * Updates a card in this address book.
   * @param aCard The card to update.
   */
  updateCard: function oAddressBook_updateCard(aCard) {
    oAbManager.checkCard(aCard);
    this.mCardsUpdate = true;
    if (this.mDirectory && this.mDirectory.modifyCard)
      this.mDirectory.modifyCard(aCard);
    else
      aCard.editCardToDatabase(this.URI);
  },
  /**
   * oAddressBook.checkList
   * Checks the validity of a mailing list and throws an error if it is invalid.
   * @param aCard        An object that should be a mailing list
   * @param aMethodName  The name of the method calling checkList (used when
   *                     throwing the error)
   */
  checkList: function oAddressBook_checkList(aList, aMethodName) {
    // if it is a oMailList object, get it's actual list
    var list = aList && aList.mList ? aList.mList : aList;
    if (!list || !(list instanceof Ci.nsIAbDirectory) || !list.isMailList) {
      throw "Invalid list: " + aList + " sent to the '" + aMethodName +
            "' method" +  oStringBundle.getStr("pleaseReport");
    }
  },
  /**
   * oAddressBook.checkDirectory
   * Checks the validity of a directory and throws an error if it is invalid.
   * @param aDirectory The directory to check.
   * @param aMethodName  The name of the method calling checkDirectory (used when
   *                     throwing the error)
   */
  checkDirectory: function oAddressBook_checkDirectory(aDirectory, aMethodName) {
    if (!this.isDirectoryValid(aDirectory))
      throw "Invalid Directory: " + aDirectory + " sent to the '" + aMethodName
            + "' method" +  oStringBundle.getStr("pleaseReport");
  },
  /**
   * oAddressBook.checkDirectory
   * Checks the validity of a directory and returns false if it is invalid.
   * @param aDirectory The directory to check.
   */
  isDirectoryValid: function oAddressBook_isDirectoryValid(aDirectory) {
    return aDirectory && aDirectory instanceof Ci.nsIAbDirectory 
          && aDirectory.dirName != "" && (oAbManager.mVersion == 3 || 
          aDirectory instanceof Ci.nsIAbMDBDirectory);
  },
  /**
   * oAddressBook.getCardValue
   * Returns the value of the specifiec property in the given card, or throws an
   * error if it is not present or blank.
   * @param aCard     The card to get the value from.
   * @param aAttrName The name of the attribute to get.
   */
   getCardValue: function oAddressBook_getCardValue(aCard, aAttrName) {
     return oAbManager.getCardValue(aCard, aAttrName);
   },
  /**
   * oAddressBook.getCardValue
   * Sets the value of the specifiec property in the given card but does not
   * update the card in the database.
   * @param aCard     The card to get the value from.
   * @param aAttrName The name of the attribute to set.
   * @param aValue    The value to set for the attribute.
   */
   setCardValue: function oAddressBook_setCardValue(aCard, aAttrName, aValue) {
     oAbManager.setCardValue(aCard, aAttrName, aValue);
   },
  /**
   * oAddressBook.hasAddress
   * Returns true if the given card has at least one address-related property
   * for the given type.
   * @param aCard    The card to check.
   * @param aPrefix  The prefix/type (Home or Work).
   * @return True if aCard has at least one address-related property of the
   *         given type.
   */
  hasAddress: function oAddressBook_hasAddress(aCard, aPrefix) {
    oAbManager.checkCard(aCard);
    return this.getCardValue(aCard, aPrefix + "Address") || 
          this.getCardValue(aCard, aPrefix + "Address2") ||
          this.getCardValue(aCard, aPrefix + "City") ||
          this.getCardValue(aCard, aPrefix + "State") ||
          this.getCardValue(aCard, aPrefix + "ZipCode") ||
          this.getCardValue(aCard, aPrefix + "Country");
  },
  /**
   * oAddressBook.makeCard
   * Creates and returns a new address book card.
   * @return A new instantiation of nsIAbCard.
   */
  makeCard: function oAddressBook_makeCard() {
    return Cc["@mozilla.org/addressbook/cardproperty;1"]
           .createInstance(Ci.nsIAbCard);
  },
  /**
   * oAddressBook.equals
   * Returns true if the directory passed in is the same as the directory
   * stored by this oAddressBook object.  Two directories are considered the same
   * if and only if their Uniform Resource Identifiers (URIs) are the same.
   * @param aOtherDir The directory to compare with this object's directory.
   * @return True if the URI of the passed directory is the same as the URI of
   *         the directory stored by this object.
   */
  equals: function oAddressBook_equals(aOtherDir) {
    // return false if the directory isn't valid
    if (!this.isDirectoryValid(aOtherDir))
      return false;
    // compare the URIs
    if (this.mDirectory.URI)
      return this.mDirectory.URI == aOtherDir.URI;
    return this.mDirectory.getDirUri() == aOtherDir.getDirUri();
  },
  /**
   * oAddressBook.hasCard
   * Returns the card in this directory, if any, with the same (not-null)
   * value for the GoogleID attribute, or, if the GoogleID is null, if the
   *         display name, primary, and second emails are the same.
   * @param aCard The card being searched for.
   * @return The card in this list, if any, with the same, and non-null value
   *         for its GoogleID attribute, or, if the GoogleID is null, if the
   *         display name, primary, and second emails are the same.
   */
  hasCard: function oAddressBook_hasCard(aCard) {
    oAbManager.checkCard(aCard);
    if (this.mCardsUpdate)
      this.getAllCards();
    for (var i = 0, length = this.mCards.length; i < length; i++) {
      var card = this.mCards[i];
      var aCardID = oAbManager.getCardValue(aCard, "GoogleID");
      // if it is an old card (has id) compare IDs
      if (aCardID) {
        if (aCardID == oAbManager.getCardValue(card, "GoogleID"))
          return card;
      }
      // else check that display name, primary and second email are equal
      else if (oAbManager.getCardValue(aCard, "DisplayName") ==
                                      oAbManager.getCardValue(card,"DisplayName")
              && oAbManager.getCardValue(aCard, "PrimaryEmail") ==
                                        oAbManager.getCardValue(card, "PrimaryEmail")
              && oAbManager.getCardValue(aCard, "SecondEmail") ==
                                        oAbManager.getCardValue(card, "SecondEmail"))
        return card;
    }
  },
  /**
   * oAddressBook.setPrefId
   * Sets the preference id for this mailing list.  The update method must be
   * called in order for the change to become permanent.
   * @param aPrefId The new preference ID for this mailing list.
   */
  setPrefId: function oAddressBook_setPrefId(aPrefId) {
    this.mDirectory.dirPrefId = aPrefId;
  },
  /**
   * oAddressBook.getPrefId
   * Returns the preference ID of this directory.
   * @return The preference ID of this directory.
   */
  getPrefId: function oAddressBook_getPrefId() {
    return this.mDirectory.dirPrefId;
  },
  /**
   * oAddressBook.getStringPref
   * Gets and returns the string preference, if possible, with the given name.
   * Returns null if this list doesn't have a preference ID or if there was an
   * error getting the preference.
   * @param aName         The name of the preference to get.
   * @param aDefaultValue The value to set the preference at if it fails.  Only
   *                      used in Thunderbird 3.
   * @return The value of the preference with the given name in the preference
   *         branch specified by the preference ID, if possible.  Otherwise null.
   */
  getStringPref: function oAddressBook_getStringPref(aName, aDefaultValue) {
    var id = this.getPrefId();
    oLOGGER.VERBOSE_LOG("Getting pref named: " + aName + " from the branch: " + id);
    /* The code below is commented out for backward compatibility with TB 2,
     * which crashes if you set a custom pref for a directory.  It is a somewhat
     * sloppy workaround that, instead of using preferences from the directory's
     * actual branch, uses a preference with from the directory's pref. ID
     * and appends the preference name to that ID without a period in between
     * ex. "ldap_2.servers.emailaddrgmailcomopenCRXPrimary" instead of
     * "ldap_2.servers.emailaddrgmailcom.openCRXPrimary" 
     */
    /*
    if (this.mDirectory.getStringValue) {
      try {
        var value = this.mDirectory.getStringValue(aName, aDefaultValue);
        oLOGGER.VERBOSE_LOG("-Found the value: " + value);
        return value;
      } catch (e) { oLOGGER.LOG_WARNING("Error while setting directory pref", e); }
      return null;
    }*/
    if (!id)
      return;
    try {
      var branch = Cc["@mozilla.org/preferences-service;1"]
                    .getService(Ci.nsIPrefService)
                    .getBranch(id)
                    .QueryInterface(Ci.nsIPrefBranch2);
      var value = branch.getCharPref(aName);
      oLOGGER.VERBOSE_LOG("-Found the value: " + value);
      return value;
    }
    catch(e) {
      oLOGGER.VERBOSE_LOG("getStringPref: (this error is usually expected)\n" + e);
    } // an error is expected if the value isn't present
  },
  /**
   * oAddressBook.setStringPref
   * Setshe string preference, if possible, with the given name and value.
   * @param aName  The name of the preference to get.
   * @param aValue The value to set the preference to.
   */
  setStringPref: function oAddressBook_setStringPref(aName, aValue) {
    var id = this.getPrefId();
    oLOGGER.VERBOSE_LOG("Setting pref named: " + aName + " to value: " + aValue +
                       " to the branch: " + id);
    /* The code below is commented out for backward compatibility with TB 2,
     * which crashes if you set a custom pref for a directory.  It is a somewhat
     * sloppy workaround that, instead of using preferences from the directory's
     * actual branch, uses a preference with from the directory's pref. ID
     * and appends the preference name to that ID without a period in between
     * ex. "ldap_2.servers.emailaddrgmailcomopenCRXPrimary" instead of
     * "ldap_2.servers.emailaddrgmailcom.openCRXPrimary"
     */
    /*
    if (this.mDirectory.setStringValue) {
      try {
        this.mDirectory.setStringValue(aName, aValue);
      } catch (e) { oLOGGER.LOG_WARNING("Error while setting directory pref", e); }
      return;
    }*/
    if (!id) {
      oLOGGER.VERBOSE_LOG("Invalid ID");
      return;
    }
    if (!aName || aName == "") {
      oLOGGER.VERBOSE_LOG("Invalid name");
      return;
    }
    try {
      var branch = Cc["@mozilla.org/preferences-service;1"]
                    .getService(Ci.nsIPrefService)
                    .getBranch(id)
                    .QueryInterface(Ci.nsIPrefBranch2);
      branch.setCharPref(aName, aValue);
    } catch(e) { oLOGGER.LOG_WARNING("Error while setting directory pref", e); }
  },
  /**
   * oAddressBook.setUrl
   * Sets the username for the account with which this address book is synced.
   * @param aUsername The username for the Google account.
   */
  setUrl: function oAddressBook_setUrl(aUrl) {
    this.setStringPref("openCRXSyncUrl", aUrl);
  },
  /**
   * oAddressBook.getUrl
   * Returns the username for the account with which this address book is synced.
   * @return The username for the Google account.
   */
  getUrl: function oAddressBook_getUrl() {
    return this.getStringPref("openCRXSyncUrl");
  },
  /**
   * oAddressBook.getName
   * Returns the name of this address book.
   * @return The name of this address book.
   */
  getName: function oAddressBook_getName() {
    return this.mDirectory.dirName;
  },
  /**
   * oAddressBook.setName
   * Sets the name of this address book.  Throws an error if the name is set to
   * either the PAB or CAB's name.
   * @param aName The new name for this directory.
   */
  setName: function oAddressBook_setName(aName) {
    // make sure it isn't being set to the PAB or CAB name and make sure that
    // this isn't the PAB or CAB
    var pab = oAbManager.getAbByURI("moz-abmdbdirectory://abook.mab");
    var cab = oAbManager.getAbByURI("moz-abmdbdirectory://history.mab");
    if (aName == pab.dirName || aName == cab.dirName)
      throw "Error - cannot rename a directory to the PAB or CAB's name";
    if (this.getName() == pab.dirName || this.getName() == cab.dirName)
      throw "Error - cannot rename the PAB or CAB";
    // in TB 3, it is as simple as changing a property of the directory
    if (oAbManager.mVersion == 3)
      this.mDirectory.dirName = aName;
    // in TB 2 a few extra steps are necessary...
    else {
      /* NOTE: this code is originally from
      * mailnews/addrbook/resources/content/addressbook.js:
      * http://mxr.mozilla.org/mozilla1.8/source/mailnews/addrbook/resources/content/addressbook.js#353
      */
      var addressbook = Cc["@mozilla.org/addressbook;1"]
                         .createInstance(Ci.nsIAddressBook);
      // the rdf service
      var RDF = Cc["@mozilla.org/rdf/rdf-service;1"].getService(Ci.nsIRDFService);
      // get the datasource for the addressdirectory
      var datasource = RDF.GetDataSource("rdf:addressdirectory");

      // moz-abdirectory:// is the RDF root to get all types of addressbooks.
      var parent = RDF.GetResource("moz-abdirectory://")
                         .QueryInterface(Ci.nsIAbDirectory);
      // Copy existing dir type category id and mod time so they won't get reset.
      var properties = this.mDirectory.directoryProperties;
      properties.description = aName;
      // Now do the modification.
      addressbook.modifyAddressBook(datasource, parent, this.mDirectory, properties);
    }
  },
   /**
    * oAddressBook.getLastSyncDate
    * Returns the last time this address book was synchronized in milliseconds
    * since the epoch.
    * @return The last time this address book was synchronized.
    */
   getLastSyncDate: function oAddressBook_getLastSyncDate() {
     return this.getStringPref("lastSync");
   },
   /**
    * oAddressBook.setLastSyncDate
    * Sets the last time this address book was synchronized, in milliseconds
    * since the epoch.
    * @param aLastSync The last sync time.
    */
   setLastSyncDate: function oAddressBook_setLastSyncDate(aLastSync) {
     this.setStringPref("lastSync", aLastSync);
   }
};
